/*  camserv - An internet streaming picture application
 *
 *  Copyright (C) 1999  Jon Travis (jon@dsndata.com)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sys/param.h>

#include "camconfig.h"
#include "camshm.h"
#include "video.h"
#include "sockset.h"
#include "socket.h"
#include "mainloop.h"
#include "picloop.h"
#include "filter.h"

static int Shmid;
static int CPid = 0;

static
void sighandler( int signum ){
  fprintf( stderr, "%s: Exiting on received signal: %d\n", 
	   (CPid == 0) ? "Slave" : "Master", signum );
  if( CPid ) kill( CPid, SIGINT );
  exit(1);
}
 
static
void setup_signals(){
  int i;

  for( i=0; i< NSIG; i++ ){
    if( i != SIGTSTP &&
	i != SIGCONT &&
	i != SIGSEGV && 
	i != SIGPROF &&
	i != SIGALRM )
      signal( i, sighandler );
  }
}
 
int main( int argc, char *argv[] ){
  Socket **localsocks;
  CamConfig *camcfg;
  Video_Funcs vfuncs;
  Video_Geometry vgeom;
  void *vid_device;
  char *shm_segment, tmpbuf[ 1024 ], cfg_path[ MAXPATHLEN ];
  int geomres, eres, fd;
  struct filter_st *filters;

  if( argc < 2 ){
    fprintf( stderr, "camserv v%s - by Jon Travis (jon@dsndata.com)\n", 
	     VERSION );
    fprintf( stderr, "Syntax: %s <cfg file>\n", argv[0] );
    fprintf( stderr,"Attempting to use installation default: %s/camserv.cfg\n",
	     DATDIR );
    snprintf( cfg_path, sizeof( cfg_path ), "%s/camserv.cfg", DATDIR );
    cfg_path[ sizeof( cfg_path ) - 1 ] = '\0';
  } else {
    strncpy( cfg_path, argv[ 1 ], sizeof( cfg_path ) );
    cfg_path[ sizeof( cfg_path ) - 1 ] = '\0';
  }

  if( (camcfg = camconfig_read( cfg_path )) == NULL ){
    fprintf( stderr, "Error reading configuration file \"%s\".  Exiting\n",
	     cfg_path );
    return -1;
  }

  setup_signals();

  if((filters =  filter_setup( camcfg, &eres )) == NULL &&
     eres == 1)
  {
    fprintf( stderr, "Error initializing filters!\n");
    return -1;
  }

  if( video_setup_funcs( camcfg, &vfuncs ) == -1 ){
    fprintf( stderr, "Error initializing video module. Exiting\n" );
    return -1;
  }  

  if( (vid_device = vfuncs.video_open( camcfg )) == NULL ){
    fprintf( stderr, "Error opening video device.  Exiting\n" );
    return -1;
  }

  geomres = vfuncs.video_get_geom( vid_device, &vgeom );
  if( !(geomres & VIDEO_GEOM_MAX )){
    fprintf( stderr, "Unable to get camera max geometry information!\n");
    return -1;
  }

  vfuncs.video_close( vid_device );

  if( (localsocks = socket_unix_pair( SOCK_DGRAM )) == NULL ){
    fprintf( stderr, "Error creating communication sockets between proc.s\n");
    return -1;
  }
  
  /* Setup a temp file for making our shm */
  strcpy( tmpbuf, "/tmp/CAMSERV_XXXXXX" );
  if( (fd = mkstemp( tmpbuf )) == -1 ){
    fprintf( stderr, "ACK!: Couldn't create temporary file: %s\n", tmpbuf );
    strcpy( tmpbuf, argv[ 0 ] ); /* Last resort */
  } else {
    close( fd );
  }

  if( (Shmid = shm_setup( tmpbuf,
			  /* Allocate generous ammount */
			  vgeom.max_width * vgeom.max_height * 3, 
			  &shm_segment) ) == -1 ){

    socket_unix_pair_dest( localsocks );
    return -1;
  }
  unlink( tmpbuf );

/* Start the picture taker thread */
  CPid = picture_taker( &vfuncs, shm_segment,
			vgeom.max_width * vgeom.max_height * 3, camcfg,
			localsocks[ 0 ], filters ); 
  if( CPid == -1 ){
    /* Failure setting up camerastuffs */
    fprintf( stderr, "Picture taker could not be created!\n");
    socket_unix_pair_dest( localsocks );
    return -1;
  }
    
  if( main_loop( camcfg, localsocks[ 1 ], shm_segment ) == -1 ){
    fprintf( stderr, "Main loop exited abnormally\n");
    socket_unix_pair_dest( localsocks );
    if( CPid != -1 ) kill( CPid, SIGINT );
    return -1;
  }

  socket_unix_pair_dest( localsocks );
  return 0;
}
  
