#include <stdio.h>
#include <stdlib.h>
#include <glib.h>
#include <string.h>

#include "camconfig.h"
#include "socket.h"
#include "mainloop.h"
#include "sockset.h"
#include "manager.h"
#include "databuf.h"

extern int errno;

#define CINFO_STATE_PREAMBLE    0
#define CINFO_STATE_PICTURE     1
#define CINFO_STATE_SEPERATOR   2
#define CINFO_STATE_SENDSIZE    3
#define CINFO_STATE_UNINIT      4

#define CLIENT_T_PROXY          0
#define CLIENT_T_BROWSER        1
#define CLIENT_T_UNINIT         2

#define RANDOMSTRING "ThisRandomString"

typedef struct client_info_st {
  Socket *socket;           
  DataBuf *writebuf;
  void *management_data;
  int state;                    /* One of CINFO_STATE_* */

  /********** Proxy only *********/
  unsigned long proxypic_size;  /* network order size of picture to send */
  char *proxypic_data;
  /*******************************/

  int last_picture_id;
  int client_type;              /* One of CLIENT_T_*    */
} ClientInfo;

static
char *get_preamble_text( size_t *len){
#define PREAMBLE_STR "HTTP/1.0 200 OK\n"  \
  "Content-type: multipart/x-mixed-replace;boundary=" RANDOMSTRING "\n" \
  "Pragma: no-cache\n\n"
  
  if( len != NULL ) *len = sizeof( PREAMBLE_STR ) - 1;
  return PREAMBLE_STR;
}

static
char *get_seperator_text( size_t *len){
#define SEPERATOR_TEXT "\n--" RANDOMSTRING "\n"  \
  "Content-type: image/jpeg\n\n" /* XXX */

  if( len != NULL ) *len = sizeof( SEPERATOR_TEXT ) - 1;
  return SEPERATOR_TEXT;
}

static 
ClientInfo *clientinfo_new( Socket *sock ){
  ClientInfo *res;
  char *preamble;
  size_t pre_size;

  if( (res = malloc( sizeof( *res ))) == NULL )
    return NULL;

  if( (res->writebuf = databuf_new()) == NULL ){
    free( res );
    return NULL;
  }

  res->management_data = NULL;
  res->socket = sock;
  res->state = CINFO_STATE_PREAMBLE;
  res->last_picture_id = -1;
  res->client_type = CLIENT_T_UNINIT;

  preamble = get_preamble_text( &pre_size );
  databuf_buf_set( res->writebuf, preamble, pre_size );
		   
  return res;
}

static
void clientinfo_dest( ClientInfo *cinfo ){
  databuf_dest( cinfo->writebuf );
  socket_dest( cinfo->socket );
  if( cinfo->management_data ){
    manager_dest_client( cinfo->management_data );
  }
  free( cinfo );
}

/*
 * accept_client:  Accept a client from the listen_socket, and return
 *                 a clientinfo structure containing it.  The new socket
 *                 will be set to non-blocking IO mode.
 *
 * Arguments:      listen_socket = Socket requiring an accept() operation.
 *
 * Return values:  Returns NULL on failure, else a valid pointer to a freshly
 *                 allocated ClientInfo structure.
 */

static
ClientInfo *accept_client( const Socket *listen_socket ){
  Socket *new_sock;
  ClientInfo *res;

  if( (new_sock = socket_accept( listen_socket )) == NULL ){
    fprintf( stderr, "Could not accept new client socket: %s\n",
	     strerror( errno ));
    return NULL;
  }

  printf( "Accepted new socket: %s\n", socket_query_remote_name( new_sock ));

  if( fcntl( socket_query_fd( new_sock ), F_SETFL, O_NONBLOCK ) == -1){
    fprintf( stderr, " Unable to set socket to nonblocking mode!\n");
    socket_dest( new_sock );
    return NULL;
  }
  
  if( (res = clientinfo_new( new_sock )) == NULL ){
    fprintf( stderr, "Error creating clientinfo structure!\n");
    socket_dest( new_sock );
    return NULL;
  }

  return res;
}
  
static
int dispatch_pictaker( const char *cmdbuf, const char *picture_mem ){
  int command;

  if( sscanf( cmdbuf, "%d", &command ) != 1 ){
    fprintf( stderr, "Parse error reading data from pictaker!\n");
    return -1;
  }

  if( command == 0 ) {
    char *new_picture_data;
    int nbytes;

    sscanf( cmdbuf, "%*d %d", &nbytes );
    if( (new_picture_data = malloc( nbytes )) == NULL ){
      fprintf( stderr, "Error allocating memory to manage picture!\n");
      return -1;
    }

    memcpy( new_picture_data, picture_mem, nbytes );
    if( manager_new_picture( new_picture_data, nbytes, 100 ) == -1 ){
      fprintf( stderr, "Unable to manage picture!\n");
      free( new_picture_data );
      return -1;
    }
  } else {
    fprintf( stderr, "Unknown pictaker dispatch command: %d\n", command );
    return -1;
  }
  return 0;
}

static
int write_proxy_client( ClientInfo *cinfo, SockSet *writeset ){
  int bufwrite;

  bufwrite = databuf_write( cinfo->writebuf, socket_query_fd( cinfo->socket ));
  if( bufwrite == -1 ) return -1;
  if( bufwrite == 1 ) return 0;

  /* In this case, we are done sending whatever it was that
     'state' referred to */
  if( cinfo->state == CINFO_STATE_SENDSIZE )
  {
    printf("Writing picture to client\n");
    databuf_buf_set( cinfo->writebuf, cinfo->proxypic_data,
		     ntohl( cinfo->proxypic_size ));
    cinfo->state = CINFO_STATE_PICTURE;
  } else { /* Just finished sending a picture */
    char *pic_data;
    size_t pic_size;
    int pic_id;
    
    /* Remove old info */
    if( cinfo->management_data != NULL ) {
      if( manager_dest_client( cinfo->management_data ) == -1 ){
	fprintf( stderr, "Error destroying client management!\n");
      }
    }

    cinfo->management_data = manager_new_client(&pic_data, &pic_size, &pic_id);
    if( cinfo->management_data == NULL ){
      fprintf( stderr, "Error managing proxy!  (Picture may not be "
	       "taken yet)\n");
      sockset_hold( writeset, cinfo->socket );
      return 0;
    }

    if( pic_id == cinfo->last_picture_id ) {
      manager_dest_client( cinfo->management_data );
      cinfo->management_data = NULL;
      /* Proxy sucking too fast */
      sockset_hold( writeset, cinfo->socket );
      return 0;
    }

    printf("Writing picsize %d to client\n", (int) pic_size );
    cinfo->proxypic_data = pic_data;
    cinfo->proxypic_size = htonl( pic_size );
    cinfo->last_picture_id = pic_id;
    databuf_buf_set( cinfo->writebuf, &cinfo->proxypic_size,
		     sizeof( cinfo->proxypic_size ));
    cinfo->state = CINFO_STATE_SENDSIZE;
  }

  return 0;
}


static
int write_regular_client( ClientInfo *cinfo, SockSet *writeset ){
  switch(databuf_write( cinfo->writebuf, socket_query_fd( cinfo->socket )))
  {
  case -1 : /* Error */
    return -1;
    break;
  case 0:  /* All done feeding client the current data */
    if( cinfo->state == CINFO_STATE_PREAMBLE   || 
	cinfo->state == CINFO_STATE_SEPERATOR )
    {
      char *pic_data;
      size_t pic_size;
      int pic_id;
      
      cinfo->management_data = manager_new_client( &pic_data, &pic_size,
						   &pic_id );
      if( cinfo->management_data == NULL ){
	fprintf( stderr, "Error managing client! (Picture may not"
		 " be taken yet)\n");
	/* Wait for the next successful picture to come around */
	sockset_hold( writeset, cinfo->socket );
	return 0;
      }
      
      if( pic_id == cinfo->last_picture_id ) {
	manager_dest_client( cinfo->management_data );
	cinfo->management_data = NULL;
	/* Whoa boy!  Hold on a second! */
	sockset_hold( writeset, cinfo->socket );
	return 0;
      }
      
      databuf_buf_set( cinfo->writebuf, pic_data, pic_size );
      cinfo->last_picture_id = pic_id;
      cinfo->state = CINFO_STATE_PICTURE;
    } else {
      /* Just finished sending a picture */
      char *sep_data;
      size_t sep_size;
      if( manager_dest_client( cinfo->management_data ) == -1 ){
	fprintf( stderr, "Error destroying client management!\n");
      }
      cinfo->management_data = NULL;
      cinfo->state = CINFO_STATE_SEPERATOR;
      sep_data = get_seperator_text( &sep_size );
      databuf_buf_set( cinfo->writebuf, sep_data, sep_size );
    }
    break;
  case 1:  /* Keep feeding data to the client */
    break;
  }
  return 0;
}


int main_loop( CamConfig *ccfg, Socket *picture_sock, char *picture_mem  ){
  Socket *listen_socket;
  SockSet *readset = NULL, *writeset = NULL;
  GList *client_sockets = NULL, *node;
  int cfg_listen_port, highest_fd, num_clients, picture_client_ready;
  ClientInfo *clientinfo, *clientinfo2;

  cfg_listen_port = camconfig_query_def_int( ccfg, "sock_listen_port",
					     CAMCONFIG_DEF_LISTEN_PORT );

  if( (readset = sockset_new()) == NULL ||
      (writeset = sockset_new()) == NULL )
  {
      fprintf( stderr, "Error allocating memory for socksets!\n");
      if( readset ) sockset_dest( readset );
      if( writeset ) sockset_dest( writeset );
      return -1;
  }

  if((listen_socket = socket_serve_tcp( NULL, cfg_listen_port, 100 )) == NULL )
  {
      fprintf( stderr, "Error setting up socket on port \"%d\".  Exiting\n",
	       cfg_listen_port  );
      return -1;
  }

  highest_fd = MAX( socket_query_fd( listen_socket ), 
		    socket_query_fd( picture_sock ));
  clientinfo = clientinfo_new( listen_socket );
  clientinfo2 = clientinfo_new( picture_sock );

  if( !clientinfo || !clientinfo2 ||
      sockset_add_fd( readset, listen_socket, clientinfo ) == -1 ||
      sockset_add_fd( readset, picture_sock, clientinfo2 ) == -1 )
  {
    fprintf( stderr, "Error adding initial sockets to sockset!\n");
    sockset_dest( readset );
    sockset_dest( writeset );
    if( clientinfo )  clientinfo_dest( clientinfo );
    if( clientinfo2 ) clientinfo_dest( clientinfo2 );
    return -1;
  }

  num_clients = 0;
  picture_client_ready = 1;

  while( 1 ){
    int sel_res, i, nset_socks;
    void **set_socks;

    if( num_clients > 0 && picture_client_ready == 1 ){
      send( socket_query_fd( picture_sock ), "0", sizeof( "0" ), 0 );
      picture_client_ready = 0;
    }

    sockset_reset( readset );
    sockset_reset( writeset );
    sel_res = sockset_select( highest_fd + 1, readset, writeset, NULL );
    /* Service the event */
    if( sel_res == -1 ){
      fprintf( stderr, "select() failure: %s\n", strerror( errno ));
      break;
    } else if( sel_res == 0 ){
      fprintf( stderr, "Unexpected select() fall through!\n" );
      continue;
    } 

    /* Readable sockets */
    set_socks = sockset_query_socks( readset );
    nset_socks = sockset_query_nsocks( readset );
    for( i=0; i< nset_socks; i++ ){
      ClientInfo *new_cinfo;

      clientinfo = set_socks[ i ];

      if( clientinfo->socket == listen_socket ) {
	/* New client */

	if( (new_cinfo = accept_client( listen_socket )) == NULL )
	  continue;

	if( sockset_add_fd( readset, new_cinfo->socket, new_cinfo ) == -1 ){
	  fprintf( stderr, "Failed to add socket %d to socket read set!\n",
		   socket_query_fd( new_cinfo->socket ));
	  clientinfo_dest( new_cinfo );
	  continue;
	}

	if( socket_query_fd( new_cinfo->socket ) > highest_fd )
	  highest_fd = socket_query_fd( new_cinfo->socket );

	client_sockets = g_list_append( client_sockets, new_cinfo );
	num_clients++;
      } else {
	char cmdbuf[ 1024 ];
	int readlen;

	clientinfo = set_socks[ i ];

	/* Regular joe client, set readable */
	if( (readlen = read( socket_query_fd( clientinfo->socket), cmdbuf, 
			     sizeof( cmdbuf ) - 1)) <= 0 )
	{
	  printf("Closing socket :%s\n", 
		 socket_query_remote_name( clientinfo->socket ));

	  client_sockets = g_list_remove( client_sockets, clientinfo );
	  sockset_del_fd( readset, clientinfo->socket );
	  sockset_unhold_all( writeset );
	  sockset_del_fd( writeset, clientinfo->socket );
	  clientinfo_dest( clientinfo );
	  num_clients--;
	} else {
	  if( clientinfo->socket == picture_sock ) {
	    if( dispatch_pictaker( cmdbuf, picture_mem ) == -1 )
	      fprintf( stderr, "Pictaker dispatch failure!\n");
	    sockset_unhold_all( writeset );
	    picture_client_ready = 1;
	  } else {
	    /* Information from a regular client */
	    cmdbuf[ readlen ] = '\0';
	    if( clientinfo->client_type == CLIENT_T_UNINIT ) {
	      /* Figure out what type of client we have */
	      if( !strncmp( cmdbuf, "GET", 3 )) {
		clientinfo->client_type = CLIENT_T_BROWSER;
	      } else if( !strncmp( cmdbuf, "PROXY", 5 )) {
		clientinfo->client_type = CLIENT_T_PROXY;
		/* Here we are in the same state as being done writing a pic */
		clientinfo->state       = CINFO_STATE_PICTURE;
		databuf_buf_set( clientinfo->writebuf, NULL, 0 ); 
	      } else 
		clientinfo->client_type = CLIENT_T_BROWSER;
	      if( sockset_add_fd( writeset, clientinfo->socket, 
				  clientinfo ) == -1 )
	      {
  		  fprintf( stderr, "Failed to add socket %d to write set!\n",
			   socket_query_fd( clientinfo->socket ));
	      }
	    }
	  }
	}
      }
    }

    if( set_socks != NULL ) free( set_socks );

    /* Writable sockets */
    set_socks = sockset_query_socks( writeset );
    nset_socks = sockset_query_nsocks( writeset );
    for( i=0; i< nset_socks; i++ ){
      ClientInfo *cinfo;

      cinfo = set_socks[ i ];
      if( cinfo->client_type == CLIENT_T_BROWSER ) {
	if( write_regular_client( cinfo, writeset ) == -1 ){
	  fprintf( stderr, "Databuf write error on socket: %d\n",
		   socket_query_fd( cinfo->socket ));
	  client_sockets = g_list_remove( client_sockets, cinfo );
	  sockset_del_fd( readset, cinfo->socket );
	  sockset_del_fd( writeset, cinfo->socket );
	  clientinfo_dest( cinfo );
	  num_clients--;
	}
      } else {
	if( write_proxy_client( cinfo, writeset ) == -1 ){
	  fprintf( stderr, "Databuf write error on socket: %d\n",
		   socket_query_fd( cinfo->socket ));
	  client_sockets = g_list_remove( client_sockets, cinfo );
	  sockset_del_fd( readset, cinfo->socket );
	  sockset_del_fd( writeset, cinfo->socket );
	  clientinfo_dest( cinfo );
	  num_clients--;
	}
      }
    }
    if( set_socks != NULL ) free( set_socks );
  }

  sockset_dest( readset );
  sockset_dest( writeset );

  for( node = client_sockets; node; node=node->next ){
    clientinfo_dest( node->data );
  }

  g_list_free( client_sockets );
  socket_dest( listen_socket );
  return 0;
}
