#include <stdio.h>
#include <stdlib.h>
#include <dlfcn.h>
#include <sys/types.h>
#include <signal.h>
#include <time.h>

#include "camconfig.h"
#include "camshm.h"
#include "video.h"
#include "sockset.h"
#include "socket.h"
#include "mainloop.h"
#include "picloop.h"
#include "filter.h"

#define MAX_CONSEC_BAD_SNAPS 5  /* Maximum # of consecutive bad snaps allowed*/
#define DEFAULT_MAX_FPS      100.0

static int initialized = 0;

static
void report_picsnap( Socket *server_sock, int bytes ){
  char buf[ 1024 ];
  
  sprintf( buf, "%d %d", 0, bytes );
  send( socket_query_fd( server_sock ),
	buf, strlen( buf ) + 1, 0 );
}

static
int init_taker( Video_Funcs *vfuncs, Filter *filters, void *vid_device,
		CamConfig *camconfig, char **memhack_buf, float *max_fps )
{
  Video_Geometry geom;
  int gres;

  if( initialized ) {
    fprintf( stderr, "BOGUS INIT TAKER DOUBLE INITIALIZATION!\n");
  }

  if( vfuncs->video_init( vid_device, camconfig ) == -1 ){
    fprintf( stderr, "Failed to initialize video device!\n");
    return -1;
  }
  
  gres = vfuncs->video_get_geom( vid_device, &geom );
  if( !(gres & VIDEO_GEOM_CUR)){
    fprintf( stderr, "Failed to get current geometry information\n");
    vfuncs->video_deinit( vid_device );
    return -1;
  }
  
  if( camconfig_query_def_int( camconfig, "video_memhack", 1 ) == 1){
    *memhack_buf = malloc( geom.cur_width * geom.cur_height * 3 );
    if( *memhack_buf == NULL ){
      fprintf( stderr, "Error allocating memhack buffer!\n");
      return -1;
    }
  } else {
    *memhack_buf = NULL;
  }
  
  *max_fps = camconfig_query_def_float( camconfig, "video_maxfps", 
					DEFAULT_MAX_FPS );
  filter_list_init( filters, camconfig );
  initialized = 1;
  return 0;
}


static
void deinit_taker( Video_Funcs *vfuncs, Filter *filters, void *vid_device,
		  CamConfig *camconfig, char **memhack_buf )
{
  if( !initialized ) {
    fprintf( stderr, "BOGUS DEINIT CALL IN PICTURE TAKER!\n");
  }

  if( vfuncs->video_deinit( vid_device ) == -1 ) 
    fprintf( stderr, "Could not de-initialize video device!\n");

  if( *memhack_buf != NULL )
    free( *memhack_buf );

  filter_list_deinit( filters );
  initialized = 0;
}    

static
void sleep_abit( float cur_fps, float max_fps )
{
  float diff;
  unsigned long msex;

  /* we don't want to make a huge chunk in the frame rate if the camera
     is going way to fast.  So we amortize it out over a few frames.  This
     isn't the greatest but it appears to work decently */

  if( cur_fps <= max_fps ) return;

  diff = cur_fps - max_fps;
  msex = (diff / max_fps) * 1000000;
  if( max_fps > .3 && msex / 1000000 > 3 ) msex = 3 * 1000000;

  sleep( msex / 1000000 );
  usleep( msex % 1000000 );
}


int picture_taker( Video_Funcs *vfuncs,
		   char *picture_memory,
		   int amt_alloced, CamConfig *camconfig, 
		   Socket *server_sock, Filter *filters ) 
{
  SockSet *sset;
  void *vid_device;
  char *memhack_buf;
  int cpid, nFrames = 0;
  time_t start_time, cur_time, last_cur_time;
  float max_fps, cur_fps;

  if( (cpid = fork()) != 0 ) {
    return cpid;
  }

  if( (vid_device = vfuncs->video_open( camconfig )) == NULL ){
    fprintf( stderr, "picloop:  Error opening video device!\n");
    kill( getppid(), SIGINT );
    return -1;
  }

  if( init_taker( vfuncs, filters, vid_device, camconfig,
		  &memhack_buf, &max_fps ) == -1 ){
    kill( getppid(), SIGINT );
    return -1;
  }

  if( (sset = sockset_new()) == NULL ){
    fprintf( stderr, "Could not create new socket set!\n");
    kill( getppid(), SIGINT );
    return -1;
  }

  if( sockset_add_fd( sset, server_sock, server_sock ) == -1 ){
    fprintf( stderr, "Could not add server socket to sockset!\n");
    sockset_dest( sset );
    kill( getppid(), SIGINT );
    return -1;
  }

  last_cur_time = 0;

  while( 1 ) {
    char buf[ 1024 ];
    int selres, dispatch_id;
    Video_Info vinfo, out_vinfo;

    sockset_reset( sset );
    selres = sockset_select( socket_query_fd( server_sock ) + 1,
			     sset, NULL, NULL );
    if( selres < 1 ) {
      fprintf( stderr, "Bad return value from select()!\n");
      sockset_dest( sset );
      kill( getppid(), SIGINT );
      return -1;
    }

    recv( socket_query_fd( server_sock ), buf, sizeof( buf ), 0 );
    if( sscanf( buf, "%d", &dispatch_id ) != 1 ){
      fprintf( stderr, "Malformed message from server!\n" );
      continue;
    }

    if( dispatch_id == 0 ){  /* Snap a picture */
      char *pic_snap;
      int nsnaps;


      for( nsnaps=0; nsnaps < MAX_CONSEC_BAD_SNAPS; nsnaps++ ){
	if( memhack_buf ) pic_snap = memhack_buf;
	else              pic_snap = picture_memory;
	if( vfuncs->video_snap( vid_device, pic_snap, &vinfo ) == -1 ){
	  fprintf( stderr, "Error snapping video image .. reiniting!\n");
	  deinit_taker( vfuncs, filters, vid_device, camconfig, 
			&memhack_buf );
	  vfuncs->video_close( vid_device );

	  /* Now reopen it and initailize it if we can */
	  if( (vid_device = vfuncs->video_open( camconfig )) == NULL ||
	      init_taker( vfuncs, filters, vid_device, camconfig, 
			  &memhack_buf, &max_fps ) == -1)
	  {
	    fprintf( stderr, "Couldn't open camera!\n");
	    if( vid_device ) vfuncs->video_close( vid_device );
	    kill( getppid(), SIGINT );
	    return -1;
	  }
	} else break;
      } 

      if( nsnaps == MAX_CONSEC_BAD_SNAPS ){
	fprintf( stderr, "Too many bad snaps!  Aborting!\n");
	deinit_taker( vfuncs, filters, vid_device, camconfig, &memhack_buf );
	vfuncs->video_close( vid_device );
	kill( getppid(), SIGINT );
	return -1;
      }

      filter_list_process( filters, pic_snap, picture_memory, 
			   &vinfo, &out_vinfo );
      report_picsnap( server_sock, out_vinfo.nbytes  );
      time( &cur_time );
      if( cur_time - last_cur_time > 10 ) /* Fudge factor */{
	time( &start_time );
	time( &cur_time );
      }
	
      if( max_fps != 0.0) {
	nFrames++;
	cur_fps = (double)nFrames / (double)(cur_time - start_time);
	last_cur_time = cur_time;
	
	fprintf( stderr, "FPS: %f (%f)\n", cur_fps, max_fps );

	if( cur_fps > max_fps ) 
	  sleep_abit( cur_fps, max_fps );
      }

    } else if( dispatch_id == 1 ) {
      init_taker( vfuncs, filters, vid_device, camconfig,
		    &memhack_buf, &max_fps );
    } else if( dispatch_id == 2 ) {
      deinit_taker( vfuncs, filters, vid_device, camconfig,
		    &memhack_buf );
    } else {
      fprintf( stderr, "Invalid command %d received from server!\n", 
	       dispatch_id );
    }
  }

  deinit_taker( vfuncs, filters, vid_device, camconfig, &memhack_buf );
  vfuncs->video_close( vid_device );
  kill( getppid(), SIGINT );
  return 0;
}

