/*  camserv - An internet streaming picture application
 *
 *  Copyright (C) 1999  Jon Travis (jon@dsndata.com)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <dlfcn.h>

#include "video.h"

/*
 * video_setup_funcs:  Open the video module, and fill out the structure
 *                     with all of the video accessing routines.
 *                     
 * Arguments:          ccfg = Camera configuration structure which
 *                            should contain VIDCONFIG_PATH which holds
 *                            the location for the video module.
 *                     vfuncs = Structure to place all of the video functions
 *                              as they are grabbed from the symbol table.
 *
 * Return values:      Returns -1 on failure (if the module could not be
 *                     opened, or a symbol could not be found), 0 on success.
 */

int video_setup_funcs( CamConfig *ccfg, Video_Funcs *vfuncs ){
  void *dlhandle;
  char *error;
  const char *module_path;

  if( (module_path = camconfig_query_str( ccfg, VIDCONFIG_PATH )) == NULL ){
      fprintf( stderr, "video: No video module path set (%s)\n",
	       VIDCONFIG_PATH );
      return -1;
  }
  if( (dlhandle = dlopen( module_path, RTLD_NOW )) == NULL ){
    fprintf( stderr, "video: %s\n", dlerror());
    return -1;
  }

  vfuncs->video_open = dlsym( dlhandle, VIDSYM_OPEN );
  if( (error = dlerror()) != NULL ){
    dlclose( dlhandle );
    fprintf( stderr, "video: %s\n", error );
    return -1;
  }

  vfuncs->video_close = dlsym( dlhandle, VIDSYM_CLOSE );
  if( (error = dlerror()) != NULL ){
    dlclose( dlhandle );
    fprintf( stderr, "video: %s\n", error );
    return -1;
  }

  vfuncs->video_init = dlsym( dlhandle, VIDSYM_INIT );
  if( (error = dlerror()) != NULL ){
    dlclose( dlhandle );
    fprintf( stderr, "video: %s\n", error );
    return -1;
  }

  vfuncs->video_deinit = dlsym( dlhandle, VIDSYM_DEINIT );
  if( (error = dlerror()) != NULL ){
    dlclose( dlhandle );
    fprintf( stderr, "video: %s\n", error );
    return -1;
  }

  vfuncs->video_snap = dlsym( dlhandle, VIDSYM_SNAP );
  if( (error = dlerror()) != NULL ){
    dlclose( dlhandle );
    fprintf( stderr, "video: %s\n", error );
    return -1;
  }

  vfuncs->video_get_geom = dlsym( dlhandle, VIDSYM_GET_GEOM );
  if( (error = dlerror()) != NULL ){
    dlclose( dlhandle );
    fprintf( stderr, "video: %s\n", error );
    return -1;
  }

  return 0;
}
