/*  camserv - An internet streaming picture application
 *
 *  Copyright (C) 1999  Jon Travis (jon@dsndata.com)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "databuf.h"
#include "socket.h"
#include "sockset.h"
#include "manager.h"

#define RELAY_BACKLOG  20

#define MAX(x,y) (((x)<(y)) ? (y) : (x))

#define CAMSOCK_STATE_UNSET     0
#define CAMSOCK_STATE_GETSIZE   1
#define CAMSOCK_STATE_GETPIC    2
#define CAMSOCK_STATE_PREAMBLE  3
#define CAMSOCK_STATE_SEPERATOR 4
#define CAMSOCK_STATE_PICTURE   5
#define CAMSOCK_STATE_SENDSIZE  6

#define CAMSOCK_CONN_UNINIT    0
#define CAMSOCK_CONN_BROWSER   1
#define CAMSOCK_CONN_PROXY     2


typedef struct client_data_st {
  Socket *cl_sock;
  DataBuf *databuf;      /* Buffer used for data sending for data going to
			    clients, and also used for receiving stuff from
			    the camera server (or other proxy) */
  void *management_data;
  int last_picture_id;

  int conn_type;
  int state;             
                         /* Only used by the connection to the camera server */
                         /* Or Proxy client connections!                     */
  unsigned long picsize; /* Picture size                                     */
  char *picbuf;          /* ^^----------^^--------^^-------------^^-------^^ */

} ClientSock;

#define RANDOMSTRING "ThisRandomString"

static
char *get_preamble_text( size_t *len){
#define PREAMBLE_STR "HTTP/1.0 200 OK\n"  \
  "Content-type: multipart/x-mixed-replace;boundary=" RANDOMSTRING "\n" \
  "Pragma: no-cache\n\n"
  
  if( len != NULL ) *len = sizeof( PREAMBLE_STR ) - 1;
  return PREAMBLE_STR;
}

static
char *get_seperator_text( size_t *len){
#define SEPERATOR_TEXT "\n--" RANDOMSTRING "\n"  \
  "Content-type: image/jpeg\n\n" /* XXX */

  if( len != NULL ) *len = sizeof( SEPERATOR_TEXT ) - 1;
  return SEPERATOR_TEXT;
}

/*
 * clientsock_new:  Create and initialize a new clientsock structure.
 *
 * Arguments:       sock = Socket to manage within the clientsock.  The
 *                         socket will be destroyed on clientsock_dest.
 *
 * Return values    Returns NULL on failure, else a valid new clientsock
 *                  structure.
 */

static
ClientSock *clientsock_new( Socket *sock ){
  ClientSock *res;

  if( (res = malloc( sizeof( *res ))) == NULL )
    return NULL;

  res->cl_sock = sock;

  if( (res->databuf = databuf_new()) == NULL ){
    free( res );
    return NULL;
  }

  res->management_data = NULL;
  res->last_picture_id = 0;
  res->conn_type = CAMSOCK_CONN_UNINIT;
  res->state     = CAMSOCK_STATE_UNSET;
  res->picsize   = 0;
  res->picbuf    = NULL;
  return res;
}

/*
 * clientsock_dest:  Destroy a clientsock structure and all of the data
 *                   stored within it.  Any socket associated with the
 *                   clientsock information will be closed and destroyed.
 *                   In addition, any databuf will be destroyed (though, the
 *                   caller must free the buffer associated with the databuf)
 *
 * Arguments:        clsock = Clientsock to destroy
 */

static
void clientsock_dest( ClientSock *clsock ){
  databuf_dest( clsock->databuf );
  socket_dest( clsock->cl_sock );
  free( clsock );
}

/*
 * process_camdata:  Process incoming data from the camera connection.
 *                   Once a new picture is received from the camera server,
 *                   it is put into the management arena.
 *
 * Arguments:        cldata = Clientdata of connection to camera server
 *                   writeset = Managed writeset from the main loop
 *
 * Return values:    Returns -1 if an error occurred, else 0
 */

static
int process_camdata( ClientSock *cldata, SockSet *writeset ){
  Socket *camsock = cldata->cl_sock;
  int readres;

  readres = databuf_read( cldata->databuf, socket_query_fd( camsock ));
  if( readres == -1 ){
    fprintf( stderr, "Error:  Failure to read from cambuf!\n");
    return -1;
  } else if( readres == 1 )  /* Still in process of reading data */
    return 0;

  if( cldata->state == CAMSOCK_STATE_GETSIZE ){
      /* Got the size of the pic -- get the actual pic now */
      cldata->picsize = ntohl( cldata->picsize );
      cldata->state = CAMSOCK_STATE_GETPIC;
      cldata->picbuf = malloc( cldata->picsize );
      if( cldata->picbuf == NULL ){
	fprintf( stderr, "Couldn't allocate %ld bytes for pic!\n",
		 cldata->picsize );
	return -1;
      }
      databuf_buf_set( cldata->databuf, cldata->picbuf,
		       cldata->picsize );
  } else {
      /* Finished receiving the picture -- commit it to the manager*/
      cldata->state = CAMSOCK_STATE_GETSIZE;
      databuf_buf_set( cldata->databuf, &cldata->picsize,
		       sizeof( cldata->picsize ));
      if( manager_new_picture( cldata->picbuf, cldata->picsize,
			       100 ) == -1 )
      {
	fprintf( stderr, "Unable to manage picture!\n");
	free( cldata->picbuf );
	return -1;
      }
      sockset_unhold_all( writeset );
  }
  return 0;
}

/*
 * accept_client:  Called when the listen sock needs to accept a client.
 *                 This routine will accept the socket, add it to the
 *                 various management sets, and place it into a clientsock.
 *
 * Arguments:      listen_sock = Listen sock
 *                 readset     = Readset as passed in by mainloop
 *                 writeset    = Writeset as passed in by main loop 
 *
 * Return values:  Returns NULL on failure, else a valid pointer to
 *                 the new clientdata on success.
 */

static
ClientSock *accept_client( const Socket *listen_sock, SockSet *readset,
		   SockSet *writeset )
{
  Socket *new_socket;
  ClientSock *res;

  if( (new_socket = socket_accept( listen_sock )) == NULL ){
    fprintf( stderr, "Error: Error accepting new client!\n");
    return NULL;
  }
  
  if( socket_set_nonblock( new_socket ) == -1 )
    fprintf( stderr, "Warning:  Couldn't set socket to NONBLOCK!\n");
  
  if( (res = clientsock_new( new_socket )) == NULL ){
    fprintf( stderr, "Error:  Error mallocing data for new client!\n");
    socket_dest( new_socket );
    return NULL;
  }
  
  if( sockset_add_fd( readset, new_socket, res ) == -1 ){
    fprintf( stderr, "Error:  Coulnd't add fd to sockset!\n");
    clientsock_dest( res );
    return NULL;
  }
  
  if( sockset_add_fd( writeset, new_socket, res ) == -1 ){
    fprintf( stderr, "Error:  Couldn't add fd to writeset!\n");
    sockset_del_fd( readset, new_socket );
    clientsock_dest( res );
    return NULL;
  }

  return res;
}

/*
 * process_client_write:  Process a client write to our server sock.  This
 *                        is really only paid attention to the very first
 *                        time to distinguish between actual clients and
 *                        proxy connections.  Once the client has established
 *                        its type, it is initialized and then all of the
 *                        sockets are unheld.
 *   
 * Arguments:             client = Client who performed a write to us
 *                        writeset = Writeset containing the client
 *
 * Return value:          Returns -1 if an error occurred, else 0
 */

static
int process_client_write( ClientSock *client, SockSet *writeset ){
  char cmdbuf[ 1024 ];
  int readlen;

  readlen = read( socket_query_fd( client->cl_sock ), cmdbuf, 
		  sizeof( cmdbuf) - 1);
  if( readlen <= 0 )
    return -1;
  
  cmdbuf[ readlen ] = '\0';

  if( !strncmp( cmdbuf, "PROXY", strlen( "PROXY"))){
    client->conn_type = CAMSOCK_CONN_PROXY;
    client->state     = CAMSOCK_STATE_PICTURE;
    databuf_buf_set( client->databuf, NULL, 0 );

  }  else {
    size_t pre_size;
    char *preamble;

    client->conn_type = CAMSOCK_CONN_BROWSER;
    client->state     = CAMSOCK_STATE_PREAMBLE;

    preamble = get_preamble_text( &pre_size );
    databuf_buf_set( client->databuf, preamble, pre_size );
    sockset_unhold_all( writeset );
  }

  return 0;
}

/*
 * service_readsocks:  The main routine for servicing all sockets whose
 *                     state has been set readable.  The camera socket,
 *                     the server socket, and all clients are checked for
 *                     readability.
 *
 * Arguments:          readset  = mainloop readset
 *                     writeset = mainloop writeset
 *                     servdata = Clientdata associated with the server socket
 *                     camdata  = Clientdata associated with the camera socket
 *                     highest_fd = Location to place the new highest_fd if
 *                                  an accepted client warrants it
 *                     clientlist = Array of connected clients.
 *
 * Return values:      Returns -1 on failure, 0 on success.
 */

static
int service_readsocks( SockSet *readset, SockSet *writeset,
		       ClientSock *servdata, ClientSock *camdata,
		       int *highest_fd,
		       ClientSock *clientlist[ FD_SETSIZE ])
{
  void **sockdata;
  int nsocks, i;
  Socket *servesock, *camsock;
  ClientSock *cldata, *new_client;

  if( (nsocks = sockset_query_nsocks( readset )) == 0) {
    return 0;
  }

  sockdata = sockset_query_socks( readset );

  servesock = servdata->cl_sock;
  camsock = camdata->cl_sock;

  for( i=0; i< nsocks; i++ ){
    cldata = sockdata[ i ];
    if( cldata == servdata ) {  /* New connection */
      if( (new_client = accept_client( servesock, readset, writeset )) != NULL)
      {
	*highest_fd = MAX( *highest_fd, socket_query_fd( new_client->cl_sock));
	fprintf( stderr, "Accepted: %s\n", 
		 socket_query_remote_name( new_client->cl_sock ));

	clientlist[ socket_query_fd( new_client->cl_sock ) ] = new_client;
      }
    } else if( cldata == camdata ) {      /* Info from the camserv */
      if( process_camdata( cldata, writeset ) == -1 ){
	free( sockdata );
	return -1;
      }
    } else {
      if( process_client_write( cldata, writeset ) == -1 ){
	int fd;

	fprintf( stderr, "Closing sock: %s\n", 
		 socket_query_remote_name( cldata->cl_sock ) );
	sockset_del_fd( readset, cldata->cl_sock );
	sockset_unhold_all( writeset );
	sockset_del_fd( writeset, cldata->cl_sock );
	fd = socket_query_fd( cldata->cl_sock );
	clientsock_dest( cldata );
	clientlist[ fd  ] = NULL;
      } 
    }
  }

  free( sockdata );
  return 0;
}

/*
 * write_proxy:  Called when a proxy client has been set writeable.
 *               This routine ensures that the right camera data and size
 *               data is sent to the proxy
 *
 * Arguments:    cldata = Clientdata of connected proxy
 *               writeset = Writeset containing cldata
 *
 * Return values:  Returns -1 on error, 0 on success.
 */

static
int write_proxy( ClientSock *cldata, SockSet *writeset ){
  Socket *proxsock;
  int writres;

  proxsock = cldata->cl_sock;
  writres = databuf_write( cldata->databuf, socket_query_fd( proxsock ) );
  if( writres == -1 ) return -1;  /* Error */
  if( writres == 1 )  return 0;   /* Still working on last segment */

  if( cldata->state == CAMSOCK_STATE_SENDSIZE ){
    databuf_buf_set( cldata->databuf,cldata->picbuf, ntohl( cldata->picsize ));
    cldata->state = CAMSOCK_STATE_PICTURE;
  } else { /* Just finished sending a picture */
    char *pic_data;
    size_t pic_size;
    int pic_id;
    
    /* Remove old info */
    if( cldata->management_data != NULL ) {
      if( manager_dest_client( cldata->management_data ) == -1 ){
	fprintf( stderr, "Error destroying client management!\n");
      }
    }

    cldata->management_data = manager_new_client( &pic_data, &pic_size, 
						  &pic_id );
    if( cldata->management_data == NULL ){
      fprintf( stderr, "Error managing proxy!  (Picture may not be "
	       "available yet)\n");
      sockset_hold( writeset, proxsock );
      return 0;
    }

    if( pic_id == cldata->last_picture_id ) {
      manager_dest_client( cldata->management_data );
      cldata->management_data = NULL;      /* Proxy sucking too fast */
      sockset_hold( writeset, proxsock );
      return 0;
    }

    fprintf( stderr,"Writing picture of size %d to client\n", (int) pic_size );
    cldata->picbuf = pic_data;
    cldata->picsize = htonl( pic_size );
    cldata->last_picture_id = pic_id;
    databuf_buf_set( cldata->databuf, &cldata->picsize,
		     sizeof( cldata->picsize ));
    cldata->state = CAMSOCK_STATE_SENDSIZE;
  }

  return 0;
}

/*
 * write_browser:  Write data to a connected browser client.  
 *
 * Arguments:      cldata = connected browser client info
 *                 writeset = writeset containing cldata
 *
 * Return values:  Returns -1 if an error occurred, else 0
 */

static
int write_browser( ClientSock *cldata, SockSet *writeset ){
  Socket *browsock;
  int writres;

  browsock = cldata->cl_sock;
  writres = databuf_write( cldata->databuf, socket_query_fd( browsock ));
  if( writres == 1 ) return 0;
  if( writres == -1 ) return -1; /* Error occurred */

  if( cldata->state == CAMSOCK_STATE_PREAMBLE ||
      cldata->state == CAMSOCK_STATE_SEPERATOR )
  {
    char *pic_data;
    size_t pic_size;
    int pic_id;

    cldata->management_data = manager_new_client( &pic_data, &pic_size,
						  &pic_id );
    if( cldata->management_data == NULL ){
      fprintf( stderr, "Error managing client!  No pictures may be available "
	       " yet\n");
      sockset_hold( writeset, browsock );
      return 0;
    }

    if( pic_id == cldata->last_picture_id ){ /* Client reading too fast! */
      manager_dest_client( cldata->management_data );
      cldata->management_data = NULL;
      sockset_hold( writeset, browsock );
      return 0;
    }
    
    databuf_buf_set( cldata->databuf, pic_data, pic_size );
    cldata->last_picture_id = pic_id;
    cldata->state = CAMSOCK_STATE_PICTURE;
  } else {
    /* Just finished sending a picture */
      char *sep_data;
      size_t sep_size;

      if( manager_dest_client( cldata->management_data ) == -1 ){
	fprintf( stderr, "Error destroying client management!\n");
      }
      cldata->management_data = NULL;
      cldata->state = CAMSOCK_STATE_SEPERATOR;
      sep_data = get_seperator_text( &sep_size );
      databuf_buf_set( cldata->databuf, sep_data, sep_size );
  }
  return 0;
}

/*
 * service_writesocks:  Called when a writeset is to be serviced, this routine
 *                      sends the correct information to all connected clients.
 *
 * Arguments:           writeset = Writeset to service.
 *
 * Return values:       Returns -1 on error, 0 on success.
 */

static
int service_writesocks( SockSet *writeset ){
  int nsocks, i;
  void **sockdata;
  ClientSock *cldata;

  if( (nsocks = sockset_query_nsocks( writeset )) == 0 )
    return 0;

  sockdata = sockset_query_socks( writeset );

  for( i=0; i< nsocks; i++ ){
    cldata = sockdata[ i ];
    
    if( cldata->conn_type == CAMSOCK_CONN_UNINIT ){
      /* Put them on hold until they give us something */
      sockset_hold( writeset, cldata->cl_sock );
      continue;
    } else if( cldata->conn_type == CAMSOCK_CONN_BROWSER ){
      write_browser( cldata, writeset );
    } else if( cldata->conn_type == CAMSOCK_CONN_PROXY ) {
      write_proxy( cldata, writeset );
    } else {
      fprintf( stderr, "Critical:  bad connection state detected!\n");
    }
  }
  free( sockdata );
  return 0;
}

/*
 * run_relay:  Setup all of the connection ports, and connect to the 
 *             camera server.  This routine is the main loop and will
 *             manage all of the incoming connections.
 *
 * Arguments:  localport = Port to bind for local listening
 *             remoteip  = Remote IP of camserver
 *             remoteport = Remote port of camserver
 *
 * Return values:  Returns -1 if an abnormal termination happened, else 0
 */

static
int run_relay( int localport, const char *remoteip, int remoteport ){
  Socket *servesock, *camsock;
  SockSet *readset = NULL, *writeset = NULL;
  int result = 0, highest_fd, i;
  ClientSock *servdata = NULL, *camdata = NULL;
  ClientSock *clientlist[ FD_SETSIZE ];

  for( i=0; i< FD_SETSIZE; i++)
    clientlist[ i ] = NULL;

  if( (servesock = socket_serve_tcp( NULL, localport, RELAY_BACKLOG ))==NULL ){
    fprintf( stderr, "Error setting up relay socket!\n");
    return -1;
  }

  if( (camsock = socket_connect( remoteip, remoteport )) == NULL ){
    fprintf( stderr, "Error connecting to camserv server: \"%s:%d\"\n",
	     remoteip, remoteport);
    socket_dest( servesock );
    return -1;
  }
  
  if( socket_set_nonblock( camsock ) == -1 )
    fprintf( stderr, "Warning:  Couldn't set camera socket to NONBLOCK!\n");

  if( (readset = sockset_new()) == NULL ||
      (writeset = sockset_new()) == NULL )
  {
    fprintf( stderr, "Error creating new socket set!\n");
    result = -1;
    goto exit_cleanup;
  }

  servdata = clientsock_new( servesock );
  camdata = clientsock_new( camsock );
  if( !servdata || !camdata ||
      sockset_add_fd( readset, servesock, servdata ) == -1 ||
      sockset_add_fd( readset, camsock, camdata ) == -1 )
  {
    fprintf( stderr, "Error adding FDs to socket set!\n");
    result = -1;
    goto exit_cleanup;
  }

  /* Setup the initial state of the camera connection */
  write( socket_query_fd( camsock), "PROXY", strlen( "PROXY" ));
  camdata->state = CAMSOCK_STATE_GETSIZE;
  databuf_buf_set( camdata->databuf, &camdata->picsize, 
		   sizeof( camdata->picsize ));

  highest_fd = MAX( socket_query_fd( servesock ), socket_query_fd( camsock ));

  while( 1 ) {
    int sel_res;

    sockset_reset( readset );
    sockset_reset( writeset );

    sel_res = sockset_select( highest_fd + 1, readset, writeset, NULL );
    if( sel_res == -1 ){
      perror( "sockset_select()" );
    } else if( sel_res == 0 ) {
      continue;
    } else {
      /* Service sockets */
      if( service_readsocks( readset, writeset, servdata, camdata, 
			     &highest_fd, clientlist ) == -1 ){
	result = -1;
	goto exit_cleanup;
      }
      service_writesocks( writeset );
    }
  }

 exit_cleanup:
  for( i=0; i< FD_SETSIZE; i++ )
    if( clientlist[ i ] ) clientsock_dest( clientlist[ i ] );

  if( readset )  sockset_dest( readset );
  if( writeset ) sockset_dest( writeset );

  if( !servdata ) socket_dest( servesock );
  else clientsock_dest( servdata );
  if( !camdata  ) socket_dest( camsock );
  else clientsock_dest( camdata );
  return result;
}


int main( int argc, char *argv[] ){
  int localport, remoteport;
  char remote_ip[ 1024 ];

  if( argc < 3 ) {
    fprintf( stderr, "camserv relay v%s - by Jon Travis (jon@dsndata.com)\n",
	     VERSION );
    fprintf( stderr, "Syntax: %s <port> <camserv IP> <ccamserv port>\n", 
	     argv[ 0 ]);
    fprintf( stderr, "\tThe relay will bind port <port> on the localhost\n");
    fprintf( stderr, "\tand connect to the camserv port to do the relay.\n");
    return -1;
  }

  if( sscanf( argv[1], "%d", &localport ) != 1 ) {
    fprintf( stderr, "Error:  port \"%s\" invalid!\n", argv[ 1 ] );
    return -1;
  }

  if( localport < 1024 ) {
    fprintf( stderr, "Warning:  Port numbers < 1024 shouldn't be used\n");
    fprintf( stderr, "\tfor the relay!\n");
  }

  if( sscanf( argv[3], "%d", &remoteport ) != 1 ) {
    fprintf( stderr, "Error:  camserv port \"%s\" invalid!\n", argv[3] );
    return -1;
  }

  strncpy( remote_ip, argv[2], sizeof( remote_ip ) - 1 );
  remote_ip[ sizeof( remote_ip ) - 1 ] = '\0';

  if( run_relay( localport, remote_ip, remoteport ) == -1 ){
    fprintf( stderr, "Camserv relay abnormally terminated\n");
    return -1;
  }
  return 0;
}
